local HealComm = LibStub:GetLibrary("LibHealComm-3.0", true)

local AceOO = AceLibrary("AceOO-2.0")

local Graph = AceOO.Class()

local BackdropEmptyBlack = { bgFile = "Interface\\ChatFrame\\ChatFrameBackground", tile = true, tileSize = 0, insets = { left = 0, right = 0, top = 0, bottom = 0 } }

local function FullUnitName(unit)
    local name, realm = UnitName(unit)
    
    if not name then return end
    
    if realm then
        return name.."-"..realm
    end
    
    return name
end

local PlayerName = FullUnitName("player")



function Graph.prototype:init(id, unit, db)
    Graph.super.prototype.init(self) 
    
    --initialize database
    self.db = db
    
    --self.db.hideOutsideRaid = self.db.hideOutsideRaid or false
    self.db.unit = unit or self.db.unit
    self.db.offTime = self.db.offTime or 0.5
    self.db.FrameTime = self.db.FrameTime or 5
    self.db.marks = self.db.marks or {}
    if self.db.IsVisible == nil then self.db.IsVisible = true end
    
    --init members
    self.id = id
    self.lines = {}
    self.marks = {}
    self.firstLine = nil
    self.healthDelta = 0
    self.lineStack = nil
    
    self.PixelPerHealth = 1
    
    --create frame
    self.frame = Graph:CreateGraphFrame(self)
    self.PixelPerTime = self.frame:GetWidth() / (self.db.FrameTime + self.db.offTime)
    
    local pos = self.db.offTime * self.PixelPerTime 
    local f = self.frame
    f.Axis:ClearAllPoints()
    f.Axis:SetPoint("TOPLEFT", f.GraphBody, "TOPLEFT", pos, 0)
	f.Axis:SetPoint("BOTTOMLEFT", f.GraphBody, "BOTTOMLEFT", pos, 0)
    
    --update unit id dependent things
    self:UpdateUnit()
    
    --create marks
    for _, mark in ipairs(self.db.marks) do
        self:CreateMark(mark, true)
    end
    
    --show or hide graph
    if self.db.IsVisible then
        self.frame:Show()
    else
        self.frame:Hide()
    end
end



---------------------------------
--- API for creator of graph ----
---------------------------------

----Update methods (called by graph creator)---

--[[Updates the time of the graph, called by creator of graph]]
function Graph.prototype:UpdateTime()    
    local firstLine = self.firstLine
    if not firstLine then return end
    
    local timeLeft = firstLine.endTime - GetTime()  

    if timeLeft < -self.db.offTime then 
        self:RemoveLine(firstLine)
        return
    end
    
    local pos = (timeLeft + self.db.offTime) * self.PixelPerTime    
    
    local f = self.frame    
    firstLine:ClearAllPoints()
    firstLine:SetPoint("TOPLEFT", f.GraphBody, "TOPLEFT", pos, 0)
	firstLine:SetPoint("BOTTOMLEFT", f.GraphBody, "BOTTOMLEFT", pos, 0)
end

--[[Updates the unit behind the unitid of the graph, called by creator of graph]]
function Graph.prototype:UpdateUnit()    
    self.unitName = FullUnitName(self.db.unit)
    
    if self.unitName then
        local name = UnitName(self.db.unit)
        self.frame.name:SetText(name)
        self:UpdateHealth()
    else
        self.frame.name:SetText(("<%s>"):format(self.db.unit))
        self.frame.missingBefore:SetText("")
        self.frame.missingAfter:SetText("")
    end    
end

--[[Updates the health of the unit of this graph, called by the creator of graph]]
function Graph.prototype:UpdateHealth(healthDelta)
    if not self.unitName then return end
    
    --get health data
    local maxHealth = UnitHealthMax(self.db.unit)
    local health = UnitHealth(self.db.unit)
    
    if healthDelta then
        self.healthDelta = min(self.healthDelta + healthDelta, maxHealth - health)
        health = health + self.healthDelta
    else
        self.healthDelta = 0
    end
    
    local missing = maxHealth - health
    
    --update PixelPerHealth (in case the maxHealth has changed)
    self.PixelPerHealth = self.frame.GraphBody:GetHeight() / maxHealth
    
    --update healthframe
    local HealthFrame = self.frame.Health
    local top = missing * self.PixelPerHealth
    HealthFrame:ClearAllPoints()
    HealthFrame:SetPoint("TOPLEFT", self.frame.GraphBody, "TOPLEFT", 0, -top)
	HealthFrame:SetPoint("BOTTOMRIGHT", self.frame.GraphBody, "BOTTOMRIGHT")
    
    if self.healthDelta ~= 0 then
        HealthFrame:SetBackdropColor(1, 1, 0, 0.1)
    else
        HealthFrame:SetBackdropColor(0, 1, 0, 0.1)
    end
    
    --update health display and coloring
    local ownHeal = self.lines[PlayerName]
    
    if ownHeal and ownHeal.hasAmount then
        local before, after = HealComm:UnitIncomingHealGet(self.unitName, ownHeal.endTime)
        before = before or 0
        after = after or 0
        
        local missingAfterMe = missing - (before + ownHeal.Amount.amount)
        
        if missingAfterMe > (ownHeal.Amount.amount * -0.5) then
            self:SetLineColor(ownHeal, 0, 1, 0, 1)
        else
            self:SetLineColor(ownHeal, 1.0, 0.5, 0, 1)
        end
        
        self.frame.missingAfter:SetText(floor(missingAfterMe - after))
    else
        local _, after = HealComm:UnitIncomingHealGet(self.unitName, GetTime() - 1)
        after = after or 0
        self.frame.missingAfter:SetText(floor(missing - after))
    end
    
    self.frame.missingBefore:SetText(floor(missing))
end


---Generel API methods----------------------

--[[Adds a marking line to the graph at the specified position]]
function Graph.prototype:CreateMark(time, dontAdd)
    local f = self.frame.GraphBody
    
    local mark = Graph:CreateLineFrame(self.frame)
    mark.time = time
    local pos = (time + self.db.offTime) * self.PixelPerTime
    mark:ClearAllPoints()
	mark:SetPoint("BOTTOMLEFT", f, "BOTTOMLEFT", pos, 0)
    mark:SetPoint("TOPLEFT", f, "TOPLEFT", pos, 0)
    mark:SetBackdropColor(1, 1, 1, 0.5)
    
    if not dontAdd then
        tinsert(self.db.marks, time)
    end
    tinsert(self.marks, mark)
end

--[[Changes the time span the graph displays (in seconds)]]
function Graph.prototype:SetTimeSpan(span)
    self.db.FrameTime = span
    self:OnResize()
end

--[[Returns the time span the graph display (in seconds)]]
function Graph.prototype:GetTimeSpan()
    return self.db.FrameTime
end


--[[Set hiding of graph outside of raids.]]
--[[
function Graph.prototype:SetHideOutsideRaid(hide)
    self.db.hideOutsideRaid = hide
    self:OnResize()
end
]]

--[[Returns hiding of graph outside of raids.]]
--[[
function Graph.prototype:GetHideOutsideRaid()
    return self.db.hideOutsideRaid
end
]]

--[[Changes the time span of the past the graph displays (in seconds)]]
function Graph.prototype:SetPastTimeSpan(span)
    self.db.offTime = span
    self:OnResize()
end

--[[Returns the time span of the past the graph display (in seconds)]]
function Graph.prototype:GetPastTimeSpan()
    return self.db.offTime
end

--[[Enables the graph]]
function Graph.prototype:Enable()
    if self.db.IsVisible then
        self.frame:Show()
        self:UpdateUnit()
    end
end

--[[Disables the graph]]
function Graph.prototype:Disable()
    self.frame:Hide()
end

--[[Shows the graph]]
function Graph.prototype:Show()
    self.db.IsVisible = true
    self.frame:Show()
    self:UpdateUnit()
end

--[[Guess what? ... Right!]]
function Graph.prototype:Hide()
    self.db.IsVisible = false
    self.frame:Hide()
end

--[[Adds a heal line to the graph. Not the same as "AddOrChangeLine()"!]]
function Graph.prototype:AddOrChangeHeal(healer, endTime, amount, dontUpdate)
    self:AddOrChangeLine(healer, endTime, amount, {r = 0.0, g = 1.0, b = 0.0, a = 1.0})
        
    if not dontUpdate then
        self:UpdateHealth(0)
    end
end

--[[Adds a general line to the graph.]]
function Graph.prototype:AddOrChangeLine(id, endTime, amount, color)
    local line = self.lines[id]
    
    
    if not line then
        self:Print("new line "..id)
        line = self:RetrieveLineFrame()
        line.id = id
        line.happened = false
        self.lines[id] = line
        line.endTime = endTime
        line.hasAmount = (amount ~= nil)
        if amount then
            if not line.Amount then
                line.Amount = Graph:CreateAmountFrame(line)
            end
            line.Amount:Show()
            line.Amount.amount = amount
        end
        
        self:AddToList(line)
    else
        self:Print("change line "..id)
        local reAttach = false
        
        if amount and (not line.hasAmount or amount ~= line.Amount.amount) then
            if not line.Amount then
                line.Amount = Graph:CreateAmountFrame(line)
            end
            line.Amount:Show()
            line.Amount.amount = amount
            line.hasAmount = true
            reAttach = true
        end
        
        if endTime and endTime ~= line.endTime then
            line.endTime = endTime
            self:RelocateInList(line)     
            reAttach = false
        end
        
        if reAttach then
            self:AdjustLineAmount(line)
        end
    end
    
    if color then
        self:SetLineColor(line, color.r, color.g, color.b, color.a)
    end
end

function Graph.prototype:RemoveEvent(id, succeededAmount)
    local line = self.lines[id]
    
    if not line then
        self:Print("event already removed "..id)
        return
    end
    
    line.happened = true
    self.lines[id] = nil
    
    
    if line.next then
        self:AdjustNextLineAmount(line)
    end
    
    self:AdjustLineAmount(line)
    
    if succeededAmount then 
        self:Print("event removed succeeded")
        self:UpdateHealth(succeededAmount) 
        if succeededAmount < 0 then
            self:SetLineColor(line, 1.0, 0.3, 0.3, 1)
        else
            self:SetLineColor(line, 0.3, 1, 0.3, 1)
        end
    else
        self:Print("event removed not succeeded")
        self:SetLineColor(line, 1, 1, 1, 0.2)
    end
end

--[[Removes the line with the passed id from the graph]]
function Graph.prototype:RemoveLine(line)
    if type(line) ~= "table" then
        return
    end
    
    self:Print("Removing line with id "..line.id)
    
    if self.lines[line.id] == line then
        self.lines[line.id] = nil
    end
    
    self:RemoveFromList(line)
    self:FreeLine(line)
end

--[[Removes the heal line with the passed id from the graph. 
    Not the same as "RemoveLine(id)"!]]
function Graph.prototype:RemoveHeal(healer, succeededAmount)
    self:RemoveEvent(healer, succeededAmount)
end

--[[Changes the color of the line with the passed id]]
function Graph.prototype:SetLineColor(line, r, g, b, a)
    if type(line) ~= "table" then
        local id = line
        line = self.lines[id]
    
        if not line then
            self:Print("SetLineColor - Invalid line id: "..id.."!")
            return
        end
    end
    
    line:SetBackdropColor(r, g, b, a)
    if line.hasAmount then
        line.Amount:SetBackdropColor(r, g, b, a)
    end
end

--[[Removes all lines from the graph.]]
function Graph.prototype:Clear()
    for id, line in pairs(self.lines) do
        self:RemoveLine(line)
    end
    
    self.lines = {}
end

--[[Changes the unit id this graph represents]]
function Graph.prototype:ChangeUnit(unit)
    self.db.unit = unit
    
    self:UpdateUnit()
end





----------------------------------------
--- Own Utility methods of the Graph ---
----------------------------------------

---Generel-----------------------------

function Graph.prototype:Print(msg)
    --DEFAULT_CHAT_FRAME:AddMessage("HealGraph["..self.id.."]: "..tostring(msg))
end

---Line methods---------------------------

function Graph.prototype:RetrieveLineFrame()
    local line = self.lineStack
    if not line then
        return Graph:CreateLineFrame(self.frame.GraphBody)
    end
    
    local nextLine = line.next
    
    self.lineStack = nextLine
    if nextLine then
        nextLine.prev = nil
    end
    
    line.prev = nil
    line.next = nil
    line.hasAmount = false
    
    line:Show()
    return line
end

function Graph.prototype:FreeLine(line)
    line.prev = nil
    line.next = self.lineStack
    self.lineStack = line
    line:Hide()
    line:ClearAllPoints()
    if line.Amount then
        line.Amount:Hide()
        line.Amount:ClearAllPoints()
    end
end

---line list handling---
function Graph.prototype:AddToList(line)
    local cur = self.firstLine
    local update = false
    if cur == nil or cur.endTime > line.endTime then --is only line or comes before first line
        line.prev = nil                                 --no earlier line
        line.next = cur                      --next line is current first line (can be nil)
        if cur then  --if there is a next, update its previous 
            cur.prev = line 
        end
        self.firstLine = line                           --update firstLine to this line   
        
    else
        while cur.next and cur.next.endTime <= line.endTime do --find the latest line which is still earlier than the new line
            cur = cur.next
        end
        line.prev = cur         --next earlier is curent
        line.next = cur.next    --next later is old next of current
        cur.next = line         --update new next of current to new line
        if line.next then --if there is a next line set its previous to the new line
            line.next.prev = line 
        end 
    end
    
    self:Print("readjusting line "..line:GetName())
    self:InsertLine(line)
end

function Graph.prototype:RelocateInList(line)
    self:RemoveFromList(line)
    self:AddToList(line)
end

function Graph.prototype:RemoveFromList(line)
    if line == self.firstLine then --line is first in list (this will happen often)
        self.firstLine = line.next  --update firstLine to the next line in the list
        if self.firstLine then --new first line does not has any previous
            self.firstLine.prev = nil 
            self:BypassLine(line)
        end 
    else
        line.prev.next = line.next --link previous to new next
        if line.next then   --if there is a next line
            line.next.prev = line.prev --link next to new previous
            self:BypassLine(line)
        end
    end
end


---Line positionioning---
function Graph.prototype:InsertLine(line)
    self:AdjustLineTime(line)
    self:AdjustLineAmount(line)
    
    if line.next then
        self:AdjustLineTime(line.next)
        self:AdjustNextLineAmount(line)
    end
end

function Graph.prototype:BypassLine(line)
    self:AdjustLineTime(line.next)
    self:AdjustNextLineAmount(line)
end

function Graph.prototype:AdjustLineTime(line)
    local parent, pos
    if line.prev then
        parent = line.prev
        pos = (line.endTime - line.prev.endTime) * self.PixelPerTime
    else
        parent = self.frame.GraphBody
        pos = (line.endTime - GetTime() + self.db.offTime) * self.PixelPerTime
    end
    
    line:ClearAllPoints()
    line:SetPoint("TOPLEFT", parent, "TOPLEFT", pos, 0)
    line:SetPoint("BOTTOMLEFT", parent, "BOTTOMLEFT", pos, 0)    
end

function Graph.prototype:AdjustNextLineAmount(line)
    line = line.next
    while line and ((not line.hasAmount) or line.happened) do
        line = line.next
    end
    
    if line then
        self:AdjustLineAmount(line)
    end
end

function Graph.prototype:AdjustLineAmount(line)
    local parent, pos, addTop
    local pred = line.prev
    
    if not line.hasAmount then return end
    
    if line.happened then        
        pred = nil
    else        
        if not line.hasAmount then return end
        
        pred = line.prev
        while pred and ((not pred.hasAmount) or pred.happened) do
            pred = pred.prev
        end
    end
    
    if pred then
        if pred.Amount.amount < 0 then
            addTop = false
        else
            addTop = true
        end
        parent = pred.Amount        
        pos = line.Amount.amount * self.PixelPerHealth
    else
        parent = self.frame.Health
        addTop = true
        pos = line.Amount.amount * self.PixelPerHealth
    end
    
    local posTop = pos
    local posBottom = 0
    if pos < 0 then 
        posTop = 0
        posBottom = pos
    end
    
    line.Amount:ClearAllPoints()
    if addTop then
        line.Amount:SetPoint("TOP", parent, "TOP", 0, posTop)
        line.Amount:SetPoint("BOTTOM", parent, "TOP", 0, posBottom)  
    else
        line.Amount:SetPoint("TOP", parent, "BOTTOM", 0, posTop)
        line.Amount:SetPoint("BOTTOM", parent, "BOTTOM", 0, posBottom)  
    end
    line.Amount:SetPoint("LEFT", line, "LEFT", 1, 0)
end

function Graph.prototype:ReadjustLine(line)
    self:AdjustLineTime(line)
    self:AdjustLineAmount(line)
end




function Graph.prototype:OnResize()
    self.PixelPerTime = self.frame:GetWidth() / (self.db.FrameTime + self.db.offTime)
    self:UpdateTime()
    self:UpdateHealth()
    
    local pos = self.db.offTime * self.PixelPerTime 
    local f = self.frame
    f.Axis:ClearAllPoints()
    f.Axis:SetPoint("TOPLEFT", f.GraphBody, "TOPLEFT", pos, 0)
	f.Axis:SetPoint("BOTTOMLEFT", f.GraphBody, "BOTTOMLEFT", pos, 0)
    
    f = f.GraphBody
    
    for _, mark in ipairs(self.marks) do
        local pos = (mark.time + self.db.offTime) * self.PixelPerTime
        mark:ClearAllPoints()
        mark:SetPoint("BOTTOMLEFT", f, "BOTTOMLEFT", pos, 0)
        mark:SetPoint("TOPLEFT", f, "TOPLEFT", pos, 0)
    end
end














----------------------------
--- Static Graph methods ---
----------------------------


--- static public API methods --------------------

--[[Sets the table the graphs can use as database to store settings (saved variables)]]
function Graph:SetDataBase(db)
    self.db = db
    
    if not self.db.graphs then
        self.db.graphs = {}
    end
    
    if not self.db.framePosition then
        self.db.framePosition = {}
    end
end

--[[Restores the graph with the passed id from the database.]]
function Graph:RestoreGraph(id)
    if not self.db then
        self:Print("No database set to be used! Can't restore frames.")
        return
    end
    
    if not self.db.graphs[id] then
        self:Print("RestoreGraph - No graph with id "..id"!")
        return
    end
    
    return Graph:new(id, nil, self.db.graphs[id])
end

--[[Tries to restore the graph with the passed id. If no graph-data for this
    id is stored it creates a new one for the passed unit]]
function Graph:RestoreOrCreateGraph(id, unit)
    if not self.db then
        self:Print("No database set to be used! Can't restore frames.")
        return
    end
    
    if not self.db.graphs[id] then
        return self:CreateNewGraph(id, unit)
    end
    
    return Graph:new(id, nil, self.db.graphs[id])
end

--[[Creates a new Graph with the passed id for the passed unit]]
function Graph:CreateNewGraph(id, unit)
    if not self.db then
        self:Print("No database set to be used! Can't restore frames.")
        return
    end
    
    self.db.graphs[id] = {}
    
    return Graph:new(id, unit, self.db.graphs[id])
end

--[[Removes the graph with the passed id]]
function Graph:RemoveGraph(id)
    if not self.db then
        self:Print("No database set to be used! Can't remove frames.")
        return
    end
    
    self.db.graphs[id] = nil
end



--- static non public methods ---------------

function Graph:Print(msg)
    --DEFAULT_CHAT_FRAME:AddMessage("HealGraph[core]: "..tostring(msg))
end


--- Event handling for graph-frames ---

--[[Starts moving or sizing the frame on mouse down]]
function Graph:OnMouseDown(f, button)
    if not f.isLocked and button == "LeftButton" and IsControlKeyDown() then
		f:StartMoving()
		f.isMoving = true
	end
	if not f.isLocked and button == "RightButton" and IsControlKeyDown() then
		f:StartSizing()
		f.isSizing = true
	end
end

--[[Stops moving or sizing the frame on mouse up]]
function Graph:OnMouseUp(f)
    if f.isMoving or f.isSizing then
		f:StopMovingOrSizing()
		f.isMoving = false
        if f.isSizing then
            f.Graph:OnResize()
        end
		f.isSizing = false
        
		self:SaveFramePosition(f)
	end
end

function Graph:OnHide(f)
	if f.isMoving or f.isSizing then
		f:StopMovingOrSizing()
		f.isMoving = false
		f.isSizing = false
	end
end


--- Frame creation---
function Graph:CreateGraphFrame(graph)
    
    local f = CreateFrame("Frame", graph.id.."GraphFrame", UIParent)
        
	f.Controller = self
    f.Graph = graph
	f:SetMovable(true)
	f:SetResizable(true)
	f:SetMinResize(50, 50)
	f:SetClampedToScreen(true)
	f:SetFrameStrata("BACKGROUND")

	f:EnableMouse(true)
	f:SetScript("OnMouseUp", function(f) f.Controller:OnMouseUp(f) end)
	f:SetScript("OnMouseDown", function(f, button) f.Controller:OnMouseDown(f, button) end)
    f:SetScript("OnHide", function(f) f.Controller:OnHide(f) end)
	f:SetBackdrop(BackdropEmptyBlack)
	f:SetBackdropColor(0, 0, 0, 0.90)
    
    f.GraphHead = CreateFrame("Frame", nil, f)
    f.GraphHead:SetPoint("TOPLEFT", f, "TOPLEFT");
    f.GraphHead:SetPoint("BOTTOMRIGHT", f, "TOPRIGHT", 0, -15);
    f.GraphHead:SetBackdrop(BackdropEmptyBlack)
    f.GraphHead:SetBackdropColor(0, 0, 1, 0.5)

	f.name = f.GraphHead:CreateFontString(nil, "OVERLAY")
	f.name:SetFontObject(GameFontNormal)
	f.name:SetJustifyH("CENTER")
	f.name:SetJustifyV("TOP")
	f.name:SetPoint("TOP", f.GraphHead, "TOP", 0, 0);
	f.name:SetText("Playername")
	f.name:SetHeight(15)

	f.missingBefore = f.GraphHead:CreateFontString(nil, "OVERLAY")
	f.missingBefore:SetFontObject(GameFontNormal)
	f.missingBefore:SetJustifyH("LEFT")
	f.missingBefore:SetJustifyV("TOP")
	f.missingBefore:SetPoint("TOPLEFT", f.GraphHead, "TOPLEFT", 2, 0);
	--f.missingBefore:SetWidth(60)
	f.missingBefore:SetText("-5700")

	f.missingAfter = f.GraphHead:CreateFontString(nil, "OVERLAY")
	f.missingAfter:SetFontObject(GameFontNormal)
	f.missingAfter:SetJustifyH("RIGHT")
	f.missingAfter:SetJustifyV("TOP")
	f.missingAfter:SetPoint("TOPRIGHT", f.GraphHead, "TOPRIGHT", -2, 0);
	--f.missingAfter:SetWidth(60)
	f.missingAfter:SetText("-1600")
    
    f.GraphBody = CreateFrame("Frame", nil, f)
    f.GraphBody:SetPoint("TOPLEFT", f.GraphHead, "BOTTOMLEFT");
    f.GraphBody:SetPoint("BOTTOMRIGHT", f, "BOTTOMRIGHT");
    f.GraphBody:SetBackdrop(BackdropEmptyBlack)
    f.GraphBody:SetBackdropColor(0, 0, 0, 0.0)
        
    f.Health = CreateFrame("Frame", nil, f.GraphBody)
    f.Health:SetBackdrop(BackdropEmptyBlack)
    f.Health:SetFrameStrata("LOW")
    f.Health:SetBackdropColor(0, 1, 0, 0.1)
    
    f.Axis = CreateFrame("Frame", nil, f.GraphBody)
    f.Axis:SetWidth(2)
    f.Axis:SetBackdrop(BackdropEmptyBlack)
    f.Axis:SetFrameStrata("MEDIUM")
    f.Axis:SetBackdropColor(1, 0.8, 0.35, 1)
    
    self:RestoreFramePosition(f)
    
    return f
end

local counter = 0
function Graph:CreateLineFrame(parent)
    local f = CreateFrame("Frame", "line"..counter, parent)
    counter = counter + 1
    f:SetWidth(1)
    f:SetBackdrop(BackdropEmptyBlack)
    f:SetFrameStrata("MEDIUM")
    f:SetBackdropColor(1, 1, 1, 1)
    
    return f
end

function Graph:CreateAmountFrame(parent)
    local f = CreateFrame("Frame", parent:GetName().."amount", parent)
    f:SetWidth(3)
    f:SetBackdrop(BackdropEmptyBlack)
    f:SetFrameStrata("MEDIUM")
    f:SetBackdropColor(1, 1, 1, 1)
    
    return f
end

-----frame appeareance-----

function Graph:SaveFramePosition(f)
    local settings = self.db.framePosition[f:GetName()]
    
    if not settings then
        self.db.framePosition[f:GetName()] = {}
        settings = self.db.framePosition[f:GetName()]
    end
    
	settings.left = f:GetLeft()
	settings.bottom = f:GetBottom()
	settings.width = f:GetWidth()
	settings.height = f:GetHeight()
end

function Graph:RestoreFramePosition(f)
    local settings = self.db.framePosition[f:GetName()]
    
    if not settings then
        self.db.framePosition[f:GetName()] = {left = 100, bottom = 600, width = 100, height = 80}
        settings = self.db.framePosition[f:GetName()]
    end
    
	f:ClearAllPoints()
	f:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", settings.left, settings.bottom)
	f:SetWidth(settings.width)
	f:SetHeight(settings.height)
end

HealGraphGraphClass = Graph