local MAJOR_VERSION = "LibIncomingEvents-1.0"
local MINOR_VERSION = tonumber(("$Revision: 72222 $"):match("%d+"))
 
local lib = LibStub:NewLibrary(MAJOR_VERSION, MINOR_VERSION)
if not lib then return end

local playerName = UnitName('player')

local InBattlegroundOrArena
local InRaidOrParty

local refTime = nil
local refTimestamp = nil

-----------DEBUG------------------

local runningID = 0;
local MobID = setmetatable({
}, {
	__index = function(t, i)
        local prefix = i:sub(1, 5)
    
        if prefix == "ENVIR" then 
            prefix = "ENVIRONMENTAL" 
        elseif prefix == "SPELL" and i:sub(7, 14) == "PERIODIC" then 
            prefix = "SPELL_PERIODIC" 
        end
        local tmp = runningID
        runningID = runningID + 1
		rawset(t, i, tmp)
		return tmp
	end,
}) 


---------------------------------

--[[Saves information about spells. Maybe persistant over logout...]]
local spellDataBase = {
    --[[table pattern:
    ["spellId"] = {        
        ["property"] = {    --like "castTime", "damage", "effectOnSwing", etc
            value,      --the actual value of the property
            accuracy,   --the accuracy of the value (since most properties are determined statistically)
            statisticData = {}, --stores statistical data needed to determine the value or increase its accuracy
        },  
    },
    ]]
}
--[[Saves information about mob types. Maybe persistant over logout...]]
local mobDataBase = {
    --[[table pattern:
    ["mobID"] = {
        properties = {
            ["property"] = {    --like "attackspeedMain", "attackSpeedOff", "mainDamage", "offDamage", etc
                value,      --the actual value of the property
                accuracy,   --the accuracy of the value (since most properties are determined statistically)
                statisticData = {}, --stores statistical data needed to determine the value or increase its accuracy
            },  
        },
        spells = {
            ["spellId"] = <link to spellDatabase table entry>,
        },
    },
    ]]
}

--[[Store every event we currently expect to happen in a sorted double linked list.]]
local expectedEventData = {
    --[[table pattern:
    first = <link to first event list entry>,
    ]]
    --[[pattern of event list entry:
    {
        prev = <link to previous event list entry>
        next = <link to next event list entry>
        expectedTime = <time the event is expected to happen>
        <additional data like source and destination, etc.>
    }
    ]]
}

--[[Saves temporary information on current active units. Mobs and players.]]
local unitData = {
    --[[table pattern:
    ["GUID"] = {
        name = nil, --if available
        outgoingDmg = {
            ["damage event(like swing)"] = {
                expectedTime,
                expiresAfter,
                targets = {
                    ["target GUID"] = {
                        probability = 1,
                        amount,
                    },
                },
            },
        },
        incomingDmg = {
            ["source GUID"]["event name"] = <link to damage event on source GUID table>,
        },
    },
    ]]
}

-----------------
-- Event Frame --
-----------------

lib.EventFrame = lib.EventFrame or CreateFrame("Frame")
lib.EventFrame:SetScript("OnEvent", function (this, event, ...) lib[event](lib, ...) end)


-------------------------------
-- Embed CallbackHandler-1.0 --
-------------------------------

lib.Callbacks = LibStub("CallbackHandler-1.0"):New(lib)

---------------------------------
-- Frequently Accessed Globals --
---------------------------------

local bit_band = _G.bit.band


---------------
-- Utilities --
---------------

local function unitFullName(unit)
    local name, realm = UnitName(unit)
    if (realm and realm ~= "") then
        return name .. "-" .. realm
    else
        return name
    end
end

local function isPlayer(unitFlags)
    return bit_band(unitFlags, COMBATLOG_OBJECT_CONTROL_PLAYER) == COMBATLOG_OBJECT_CONTROL_PLAYER
end

local function isMob(unitFlags)
    return  bit_band(unitFlags, COMBATLOG_OBJECT_CONTROL_NPC) == COMBATLOG_OBJECT_CONTROL_NPC and 
            bit_band(unitFlags, COMBATLOG_OBJECT_REACTION_HOSTILE) == COMBATLOG_OBJECT_REACTION_HOSTILE and 
            (bit_band(unitFlags, COMBATLOG_OBJECT_TYPE_NPC) == COMBATLOG_OBJECT_TYPE_NPC or
             bit_band(unitFlags, COMBATLOG_OBJECT_TYPE_GUARDIAN) == COMBATLOG_OBJECT_TYPE_GUARDIAN)
end

function lib:Print(msg)
    --DEFAULT_CHAT_FRAME:AddMessage("IncomingEvents: "..tostring(msg))
end

local function timestampToTime(timestamp)
    return timestamp + refTime
end

local function timeToTimestamp(time)
    return time - refTime
end

---------------------
-- Data management --
---------------------

--[[Returns the default data for a new mob type.]]
local function getMobDefault()
    return {
        properties = {
            ["attackspeedMain"] = {
                value = 2.0,        --most mobs have an attackspeed of 2.0
                accuracy = 0.4,  
                guidValue = {},
            },
            ["hitChanceMain"] = {
                value = 1.0,
                accuracy = 0.1,
                guidValue = {},
            },
            ["damageMain"] = {
                accuracy = 0.000001,
                guidValue = {},
            },
            ["attackspeedOff"] = {
                value = 2.0,        --most mobs have an attackspeed of 2.0
                accuracy = 0.4,  
                guidValue = {},
            },
            ["hitChanceOff"] = {
                value = 1.0,
                accuracy = 0.1,
                guidValue = {},
            },
            ["damageOff"] = {
                accuracy = 0.000001,
                guidValue = {},
            },
            ["dualWield"] = {
                value = false,
                accuracy = 0.5,
            }
        }, 
    }
end

--[[Returns the default data for a new spell.]]
local function spellPropertyDefaults()
    return {
        ["casttime"] = {
            value = 0.0,        --most spells will be instant, because otherwise we whould call it a cast
            accuracy = 10,  
            guidValue = {},
        },
        ["affectsPlayer"] = {
            value = true,
            accuracy = 0.000001,
        },
        ["hitChance"] = {
            value = 1.0,
            accuracy = 10.0,
            guidValue = {},
        },
        ["amount"] = {
            accuracy = 0.0000001,
            guidValue = {},
        },
    }
end


--[[Returns the default data for a new cast.]]
local function castPropertyDefaults() 
    return setmetatable({
        ["casttime"] = {
            value = 3,        --most spells will be instant, because otherwise we whould call it a cast
            accuracy = 0.05,  
            guidValue = {},
        },
    }, {
        __index = function(t, i)        
            return spellPropertyDefaults()[i]
        end,
    }) 
end

--[[Returns information on a the mob type with the passed id]]
local function GUIDToID(guid)
    return tonumber(guid:sub(-12,-7),16)
end

local function getMobData(id)
    local mob = mobDataBase[id]
    if not mob then
        mobDataBase[id] = getMobDefault()
        mob = mobDataBase[id]
    end
    
    return mob
end

local function getMobDataForGUID(guid)
    return getMobData(GUIDToID(guid))
end

local function getMobPropertyForMob(mob, property, dstGUID)
    local prop = mob.properties[property]
    
    if not prop then
        return
    end
    
    if dstGUID and prop.guidValue and prop.guidValue[dstGUID] then
        return prop.guidValue[dstGUID]
    else
        return prop.value
    end
end

local function getMobPropertyForGUID(guid, property, dstGUID)    
    return getMobPropertyForMob(getMobDataForGUID(guid), property, dstGUID)
end

local function getMobPropertyDataForMob(mob, property, createIfMissing)
    local prop = mob.properties[property]
    if not prop then
        mob.properties[property] = {accuracy = 0}
        prop = mob.properties[property]
    end
    return prop
end

local function getMobPropertyDataForGUID(guid, property, createIfMissing)    
    return getMobPropertyDataForMob(getMobDataForGUID(guid), property, createIfMissing)
end

local function getSpellData(spellId)
    --lib:Print("Getting spell with id "..spellId)
    local spell = spellDataBase[spellId]
    
    if not spell then
        spellDataBase[spellId] = {}
        spell = spellDataBase[spellId]
    end
    
    return spell
end

local function getSpellPropertyDataForSpell(spell, property)
    local prop = spell[property]
    
    if not prop then
        spell[property] = spellPropertyDefaults()[property]
        prop = spell[property]
    end
    
    return prop
end

local function getSpellPropertyDataForSpellID(spellId, property)
    return getSpellPropertyDataForSpell(getSpellData(spellId), property)    
end

local function getCastPropertyDataForSpell(spell, property)
    local prop = spell[property]
    
    if not prop then
        spell[property] = castPropertyDefaults()[property]
        prop = spell[property]
    end
    
    return prop
end

local function getCastPropertyDataForSpellID(spellId, property)
    return getCastPropertyDataForSpell(getSpellData(spellId), property)    
end

local function getSpellPropertyForSpell(spell, property, dstGUID)
    local prop = spell[property]
    
    if not prop then
        spell[property] = spellPropertyDefaults()[property]
        prop = spell[property]
    end
    
    if dstGUID and prop.guidValue and prop.guidValue[dstGUID] then
        return prop.guidValue[dstGUID]
    else
        return prop.value
    end
end

local function getSpellPropertyForSpellID(spellId, property, dstGUID)    
    return getSpellPropertyForSpell(getSpellData(spellId), property, dstGUID)
end

---------------------
-- Unit management --
---------------------

--[[Checks for events which expected time has expired a certain time.]]
local function clearOutdatedData()
    --[[
    local currentTime = GetTime() - (refTime or 0)
    
    for unitGUID, data in pairs(unitData) do
        if data.outgoingDmg then
            for eventName, eventData in pairs(data.outgoingDmg) do                
                if currentTime > eventData.startTime + eventData.expectedAfter + eventData.expiresAfter then
                    lib:Print("stopping "..eventName.." for "..unitGUID)
                    lib:EventStopped(currentTime, eventName, eventData, unitGUID, nil, false, nil)
                end
            end
        end
    end
    ]]
end

local function  addEventToUnit(srcGUID, srcUnit, dstGUID, eventName, startTime, expectedAfter, expiresAfter, amount, hitChance)
    --lib:Print("add event:"..srcGUID..", "..eventName)
    if not srcUnit then
        srcUnit = unitData[srcGUID]
        
        if not srcUnit then
            unitData[srcGUID] = {outgoingDmg = {}}
            srcUnit = unitData[srcGUID]
        end
    end
    
    local event = srcUnit.outgoingDmg[eventName]
    if not event then
        srcUnit.outgoingDmg[eventName] = {targets = {}}
        event = srcUnit.outgoingDmg[eventName]
    end
    
    event.startTime = startTime
    event.expectedAfter = expectedAfter
    event.expiresAfter = expiresAfter
    event.targets[dstGUID] = {amount = amount, probability = hitChance}
    
    local dstUnit = unitData[dstGUID]
    if not dstUnit then
        unitData[dstGUID] = {incomingDmg = {}}
        dstUnit = unitData[dstGUID]
    end
    local dstUnitInc = dstUnit.incomingDmg[srcGUID]
    if not dstUnitInc then
        dstUnit.incomingDmg[srcGUID] = {}
        dstUnitInc = dstUnit.incomingDmg[srcGUID]
    end
    dstUnitInc[eventName] = event
end

--[[Removes the passed event for the passed source and destination unit
    (or all destinations if ommitted).]]
local function removeEventFromUnit(srcGUID, unit, eventName, dstGUID)
    --lib:Print("remove event:"..srcGUID..", "..eventName..", "..(dstGUID or "nil"))
    
    if not unit then
        unit = unitData[srcGUID]
        
        if not unit then
            --lib:Print("removeEventFromUnit: unknown unit with guid "..srcGUID)
            return
        end
    end
    
    local event = unit.outgoingDmg[eventName]    
    if not event then return end
    
    if dstGUID then
        if event.targets[dstGUID] then
            local target = unitData[targetGUID]
            
            if target and target.incomingDmg and target.incomingDmg[srcGUID] then
                target.incomingDmg[srcGUID][eventName] = nil
            end
        
            event.targets[dstGUID] = nil
            
            if not next(event.targets) then
                unit.outgoingDmg[eventName] = nil
            end
        end
    else
        for targetGUID, data in pairs(event.targets) do
            local target = unitData[targetGUID]
            
            if target and target.incomingDmg and target.incomingDmg[srcGUID] then
                target.incomingDmg[srcGUID][eventName] = nil
            end
        end
        unit.outgoingDmg[eventName] = nil
    end
end

local function getUnitData(guid)
    local data = unitData[guid] 
    
    if not data then
        unitData[guid] = {outgoingDmg = {}}
        data = unitData[guid]
    end
    return data
end

local function getUnitsLastTarget(guid)
    return getUnitData(guid).lastTarget
end

local function setUnitsLastTarget(guid, targetGUID)
    --lib:Print("set last target")
    getUnitData(guid).lastTarget = targetGUID
end

local function getUnitOutgoingEvents(guid)
    local unit = unitData[guid]
    
    if not unit then return end
    
    return unit.outgoingDmg
end

local function getUnitIncomingEvents(guid)
    local unit = unitData[guid]
    
    if not unit then return end
    
    return unit.incomingDmg
end

local function getUnitOutgoingEvent(guid, eventName)
    local unit = unitData[guid]
    
    if not unit then return end
    
    return unit.outgoingDmg[eventName]
end

-----------------------
-- Damage evaluation --
-----------------------

--[[Removes the passed event for the passed source and destination unit 
    (or all destinations if ommitted) and fires a Damage_Stopped for every destination
    (or all destinations).
    "event"-field is optional.]]
function lib:EventStopped(timestamp, eventName, event, srcGUID, dstGUID, hit, amount)
    if dstGUID then
        self:Print("Fire stop.")
        self.Callbacks:Fire("IncomingEvents_Stopped", eventName, srcGUID, dstGUID, hit, amount, timestampToTime(timestamp))
    else
        if not event then
            event = getUnitOutgoingEvent(srcGUID, eventName)
        end
        if event then
            for targetGUID, data in pairs(event.targets) do
                self:Print("Fire stop.")
                self.Callbacks:Fire("IncomingEvents_Stopped", eventName, srcGUID, targetGUID, hit, amount, timestampToTime(timestamp))
            end
        end
    end
    
    removeEventFromUnit(srcGUID, nil, eventName, dstGUID)    
end

function lib:EventStarted(eventName, startTime, expectedAfter, srcGUID, dstGUID, amount, hitChance)
    addEventToUnit(srcGUID, nil, dstGUID, eventName, startTime, expectedAfter, expectedAfter * 0.5, amount, hitChance)
    self:Print("Fire start of "..eventName.." expected in "..expectedAfter)
    self.Callbacks:Fire("IncomingEvents_Started", eventName,
                        srcGUID, dstGUID, 
                        timestampToTime(startTime) + expectedAfter, 
                        amount, hitChance)
end

function lib:EventDelayed(eventName, event, newExpectedAfter, guid)
    event.expectedAfter = newExpectedAfter
    --self:Print("Delay.")
    local newTime = timestampToTime(event.startTime) + newExpectedAfter
    for targetGUID, data in pairs(event.targets) do
        --self:Print("Fire change.")
        self.Callbacks:Fire("IncomingEvents_TimeChanged", eventName, 
                            guid, targetGUID, 
                            newTime)
    end
end

function lib:DeathOfMob(timestamp, guid)
    local outgoing = getUnitOutgoingEvents(guid)
    if not outgoing then return end
    
    for eventName, event in pairs(outgoing) do
        self:EventStopped(timestamp, eventName, event, guid, nil, false)
    end
end


--------------------------
-- Statistic evaluation --
--------------------------

local function updateMedianicStat(stats, statName, data)
    local medStats = stats[statName]
    
    if not medStats then
        stats[statName] = {count = 0, sum = 0, dev = 0}
        medStats = stats[statName]
        --lib:Print(("%s: Created"):format(statName))
    end
    
    local sum = medStats.sum
    local dev = medStats.dev
    local count = medStats.count
    
    if count < 200 then   
        
        sum = sum + data
        count = count + 1
        dev = dev + abs(sum / count - data)        
        
        medStats.count = count
        medStats.sum = sum         
        medStats.dev = dev
         
        lib:Print(("%s: New median=%.2f deviance=%.2f"):format(statName, sum / count, dev / count))
    end
    
    if count < 10 then
        return sum / count, 0.1 * count
    else
        return sum / count, sum / (dev * 1.5)
    end    
end

local function addSumDataToProperty(property, data, dstGUID)   
    --lib:Print("Addsumtoprop: "..data.." - "..(dstGUID or "nil"))
    local stats = property.statisticData
    
    if not stats then
        property.statisticData = {}
        stats = property.statisticData
    end
    
    property.value, property.accuracy = updateMedianicStat(stats, "generellMed", data)
    
    if dstGUID then
        property.guidValue[dstGUID] = updateMedianicStat(stats, dstGUID.."Med", data)
    end
end

local function setStartToTimeProperty(property, time, guid)
    local stats = property.statisticData
    if not stats then
        property.statisticData = {forGUID = {}}
        stats = property.statisticData
    end
    
    local guidStats = stats.forGUID[guid]
    if not guidStats then
        stats.forGUID[guid] = {}
        guidStats = stats.forGUID[guid]
    end    
    
    guidStats.startTime = time
end

local function setLaggedStartToTimeProperty(property, time, guid)
    local stats = property.statisticData
    
    local guidStats = stats.forGUID[guid]
        
    guidStats.startTimeLagged = time
end

local function cancelTimeProperty(property, guid)
    local stats = property.statisticData
    if not stats then
        return
    end
    
    local guidStats = stats.forGUID[guid]
    if not guidStats then
        return
    end    
    
    guidStats.startTime = nil
    guidStats.delay = 0
end

local function hasTimePropertyStarted(property, guid)
    local stats = property.statisticData
    if not stats then return false end
    
    local guidStats = stats.forGUID[guid]
    if not guidStats then return false end
    
    return guidStats.startTime ~= nil
end

local function getPropertyValueTollerance(property)
    --lib:Print(("Tollerance is %.2f / %.2f"):format(property.value, property.accuracy))
    return property.value / property.accuracy
end

local function isValueOutlier(property, value, tollerance)
    lib:Print(("Outliercheck: value[%.2f], expected[%.2f], accuracy[%.2f]"):format(value, property.value, property.accuracy))
    lib:Print(("Outliercheck: delta[%.2f] < acceptableDelta[%.2f]"):format(abs(property.value - value),
                                                                           tollerance))
                                                                           
    tollerance = tollerance or getPropertyValueTollerance(property)
    return abs(property.value - value) > tollerance
end

local function setEndToTimeProperty(property, time, guid)
    local stats = property.statisticData
    if not stats then return end
    
    local guidStats = stats.forGUID[guid]
    if not guidStats then return end
    
    if not guidStats.startTime then return end
    
    if guidStats.startTimeLagged then
        if time - guidStats.startTimeLagged > 0.5 then 
            guidStats.startTime = guidStats.startTimeLagged
            guidStats.startTimeLagged = nil
        end
    end
    
    local delta = time - guidStats.startTime - (guidStats.delay or 0)
    
    --every second hit time could be nearer at the expected
    if guidStats.prevStartTime then
        local prevDelta = (time - guidStats.prevStartTime) * 0.5
        
        if abs(property.value - prevDelta) < abs(property.value - delta) then            
            lib:Print(("Previous delta(%.2f) nearer at expected than delta(%.2f)"):format(prevDelta, delta))
            delta = prevDelta
        end
    end
    
    local tollerance = getPropertyValueTollerance(property)
    local isOutlier = isValueOutlier(property, delta, tollerance)
    
    
    if not isOutlier then   
        property.value, property.accuracy = updateMedianicStat(stats, "generellMed", delta)
                
        if property.accuracy > 80 then
            property.guidValue[guid] = property.value
        else
            property.guidValue[guid] = delta
        end
        
        if guidStats.outlierSerie then
            lib:Print("Outlier serie of "..guidStats.outlierSerie.." ended")
            guidStats.outlierSerie = nil
        end
    else
        guidStats.outlierSerie = (guidStats.outlierSerie or 0) + 1
        lib:Print("Outlier")
        if guidStats.outlierSerie > 1 then
            property.guidValue[guid] = delta
        else
            property.guidValue[guid] = property.value
        end
    end
    
    guidStats.delay = 0
    guidStats.prevStartTime = guidStats.startTime
    
    if guidStats.startTimeLagged then
        guidStats.startTime = guidStats.startTimeLagged
        guidStats.startTimeLagged = nil
    else
        guidStats.startTime = nil
    end   
end

local function setDelayToTimeProperty(property, delay, guid)
    local stats = property.statisticData
    if not stats then return end
    
    local guidStats = stats.forGUID[guid]
    if not guidStats then return end    
    
    guidStats.delay = (guidStats.delay or 0) + delay
end

local function addValueToBoolProperty(property, data, guid)
    local stats = property.statisticData
    
    if not stats then
        property.statisticData = {}
        stats = property.statisticData
    end
    
    local val, accuracy = updateMedianicStat(stats, "boolMed", data and 1 or 0)
    
    property.value = val > 0.5
    property.accuracy = abs(val - 0.5) / 0.5 * 100
    
    
    if guid then
        val = updateMedianicStat(stats, guid.."Med", data and 1 or 0)
                
        property.guidValue[guid] = val > 0.5
    end
end


function lib:EvaluateReplaceSwingEnd(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool)
    local propData = getSpellPropertyData(spellId, "effectOnSwing")
    
    if propData.accuracy > 0.95 then return end
    
    local stats = propData.statisticData    
    if not stats then return end
    
    if stats.forGUID[srcGUID] then
        local nextExpectedSwing = getNextSwingFor(srcGUID)
        local delta = abs(timestamp - nextExpectedSwing)
    end
end

function lib:EvaluateReplaceSwingStart(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool)
    local propData = getSpellPropertyData(spellId, "effectOnSwing")
    
    if propData.accuracy > 0.95 then return end
    
    local stats = propData.statisticData
    
    if not stats then
        propData.statisticData = {forGUID = {}}
        stats = propData.statisticData
    end
    
    stats.forGUID[srcGUID] = true
end


------------------
-- SwingHandler --
------------------

local SwingHandler = {isAutoAttacking = {}}

function SwingHandler:SwingOnPlayer(hand, timestamp, srcGUID, srcName, dstGUID, dstName, hitType, amount, school, resisted, blocked, absorbed, critical, glancing, crushing)
    --fire damage event
    lib:EventStopped(timestamp, "swing"..hand, nil, srcGUID, dstGUID, hitType == "HIT", amount)
    setUnitsLastTarget(srcGUID, dstGUID)
    local lt = getUnitsLastTarget(srcGUID)
    --lib:Print(MobID[dstGUID].." <> "..(lt and MobID[lt] or "nil"))
    --fire next swing incoming damage event
    
    local mob = getMobDataForGUID(srcGUID)
    
    --lib:Print("expect in: "..getMobPropertyForMob(mob, "attackspeed"..hand, srcGUID))
    
    lib:EventStarted("swing"..hand, timestamp, 
                      getMobPropertyForMob(mob, "attackspeed"..hand, srcGUID), 
                      srcGUID, dstGUID, 
                      getMobPropertyForMob(mob, "damage"..hand, dstGUID), 
                      getMobPropertyForMob(mob, "hitChance"..hand, dstGUID))
end

function SwingHandler:EvaluateMobHit(hand, timestamp, eventtype, srcGUID, srcName, dstGUID, dstName, hitType)
    if hitType == "ABSORB" then return end
    
    local hitData = getMobPropertyDataForGUID(srcGUID, "hitChance"..hand)
        
    local hit = 0
    if hitType == "HIT" then
        hit = 1
    end   
    --lib:Print("Addsumtoprop: hitChanceMain")
    addSumDataToProperty(hitData, hit, dstGUID)
end

function SwingHandler:EvaluateMobDamage(hand, timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, amount, school, resisted, blocked, absorbed, critical, glancing, crushing)
    local damageData = getMobPropertyDataForGUID(srcGUID, "damage"..hand, true)
    
    amount = amount + (absorbed or 0) + (blocked or 0)
    --lib:Print("Addsumtoprop: damageMain")
    addSumDataToProperty(damageData, amount)
    
    local factor = (critical and 0.5) or (crushing and 0.6666667) or 1.0
    damageData.guidValue[dstGUID] = amount * factor
end

function SwingHandler:EvaluateMobSpeed(hand, timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags)
    local speedData = getMobPropertyDataForGUID(srcGUID, "attackspeed"..hand)
    
    setEndToTimeProperty(speedData, timestamp, srcGUID)
    setStartToTimeProperty(speedData, timestamp, srcGUID)
end


function SwingHandler:EvaluateMobParry(timestamp, guid)
    local speedData = getMobPropertyDataForGUID(guid, "attackspeedMain")
    
    lib:Print("EV parry")
    local stats = speedData.statisticData
    if not stats then return end
    
    local guidStats = stats.forGUID[guid]
    if not guidStats then return end
    
    if not guidStats.startTime then return end
    
    local speed = speedData.guidValue[guid] or speedData.value
    local expected = guidStats.startTime + speed + (guidStats.delay or 0)
    local remaining = expected - timestamp
    lib:Print(remaining.." > "..(speed * 0.2))
    if remaining > speed * 0.2 then
        local delay = (timestamp + math.max(speed * 0.2, remaining - speed * 0.4)) - expected
        lib:Print("delay: "..delay)
        setDelayToTimeProperty(speedData, delay, guid)
        
        local swingEvent = getUnitOutgoingEvent(guid, "swingMain")
        if swingEvent then 
            lib:EventDelayed("swingMain", swingEvent, speed + guidStats.delay, guid)
        end
    end
end

function SwingHandler:GuessCurrentHand(timestamp, mobGUID, dstGUID, hasHit, damage)

    local mainEvent = getUnitOutgoingEvent(mobGUID, "swingMain")
    
    if mainEvent and timestamp - (mainEvent.startTime + mainEvent.expectedAfter) > 15 then
        lib:Print("To much time since last swing -> reset.")
        self:StopAutoAttacking(timestamp, mobGUID)
        mainEvent = nil
    end
    
    if not mainEvent or not self.isAutoAttacking[mobGUID] then
        self.isAutoAttacking[mobGUID] = timestamp
        lib:Print("Autoattack started")
        
        return "Main"
    elseif self.isAutoAttacking[mobGUID] > 0 then
        local dualData = getMobPropertyDataForGUID(mobGUID, "dualWield")
        if timestamp - self.isAutoAttacking[mobGUID] < 1.4 then
            lib:Print("This could be dual wield")
            addValueToBoolProperty(dualData, true)
            self.isAutoAttacking[mobGUID] = -1
            return "Off"
        else
            addValueToBoolProperty(dualData, false)
            self.isAutoAttacking[mobGUID] = -1
            return "Main"
        end
    end
    
    local mob = getMobDataForGUID(mobGUID)
    
    if not getMobPropertyForMob(mob, "dualWield") then
        return "Main"
    end
    
    mainEvent = mainEvent or getUnitOutgoingEvent(mobGUID, "swingMain")
    local offEvent = getUnitOutgoingEvent(mobGUID, "swingOff")    
    
    lib:Print(("startMain = %.2f expAfterMain = %.2f"):format(mainEvent.startTime, mainEvent.expectedAfter))
    local mainDelta = abs(timestamp - (mainEvent.startTime + mainEvent.expectedAfter))
    local offDelta = abs(timestamp - (offEvent.startTime + offEvent.expectedAfter))
        
    lib:Print(("Offdelta=%.2f Maindelta=%.2f"):format(offDelta, mainDelta))
    if mainDelta < offDelta then
        return "Main"
    else
        return "Off"
    end
end

function SwingHandler:OnMobHit(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags,amount, school, resisted, blocked, absorbed, critical, glancing, crushing)
    local hand = self:GuessCurrentHand(timestamp, srcGUID, dstGUID, true, amount)
    
    self:EvaluateMobSpeed(hand, timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags)
    self:EvaluateMobHit(hand, timestamp, eventtype, srcGUID, srcName, dstGUID, dstName, "HIT")
    self:EvaluateMobDamage(hand, timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags,amount, school, resisted, blocked, absorbed, critical, glancing, crushing)
    
    self:SwingOnPlayer(hand, timestamp, srcGUID, srcName, dstGUID, dstName, "HIT", amount, school, resisted, blocked, absorbed, critical, glancing, crushing)
end

function SwingHandler:OnMobMiss(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags,missType)
    local hand = self:GuessCurrentHand(timestamp, srcGUID, dstGUID, false, nil)
    
    self:EvaluateMobSpeed(hand, timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags)
    self:EvaluateMobHit(hand, timestamp, eventtype, srcGUID, srcName, dstGUID, dstName, missType)
    
    self:SwingOnPlayer(hand, timestamp, srcGUID, srcName, dstGUID, dstName, missType)
end

function SwingHandler:OnMobParry(timestamp, mobGUID)
    self:EvaluateMobParry(timestamp, mobGUID)
end

function SwingHandler:StopAutoAttacking(timestamp, srcGUID)
    local speedData = getMobPropertyDataForGUID(srcGUID, "attackspeedMain")    
    cancelTimeProperty(speedData, srcGUID)
    lib:EventStopped(timestamp, "swingMain", nil, srcGUID, nil, false)
    
    speedData = getMobPropertyDataForGUID(srcGUID, "attackspeedOff")    
    cancelTimeProperty(speedData, srcGUID)    
    lib:EventStopped(timestamp, "swingOff", nil, srcGUID, nil, false)
    
    self.isAutoAttacking[srcGUID] = nil
end

function SwingHandler:OnSpellStart(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool)
    self:StopAutoAttacking(timestamp, srcGUID)
end

function SwingHandler:OnLeftCombat()
    local guid = UnitGUID("player")
    local incoming = getUnitIncomingEvents(guid)
    if not incoming then return end
    local timestamp = timeToTimestamp(GetTime())
    for srcGUID, events in pairs(incoming) do
        self:StopAutoAttacking(timestamp, srcGUID)
    end
end



------------------
-- SpellHandler --
------------------

local SpellHandler = {castingSpell = {}}

function SpellHandler:OnSpellStart(timestamp, srcGUID, spellId, spellName, spellSchool)
    local spellData = getSpellData(spellId)
    local casttimeData = getCastPropertyDataForSpell(spellData, "casttime")
    
    local currentCast = self.castingSpell[srcGUID]
    if currentCast and currentCast == spellId then
        setLaggedStartToTimeProperty(casttimeData, timestamp, srcGUID)
        lib:Print("Got spellstart before spell end. Could be lag...")
        return
    else
        setStartToTimeProperty(casttimeData, timestamp, srcGUID)
        self.castingSpell[srcGUID] = spellId
    end
    
    
    local lastTarget = getUnitsLastTarget(srcGUID)
    --lib:Print(MobID[srcGUID].." - Last target: "..(lastTarget and MobID[lastTarget] or "nil"))
    --lib:Print(getSpellPropertyForSpell(spellData, "affectsPlayer") and "affects player" or "does not affect player")
    if lastTarget and getSpellPropertyForSpell(spellData, "affectsPlayer") then
        lib:EventStarted(spellName, timestamp, 
                         getSpellPropertyForSpell(spellData, "casttime", srcGUID), 
                         srcGUID, lastTarget, 
                         getSpellPropertyForSpell(spellData, "amount", dstGUID), 
                         getSpellPropertyForSpell(spellData, "hitChance", dstGUID))
    end
end

function SpellHandler:OnSwing(timestamp, srcGUID)
    local spellId = self.castingSpell[srcGUID]
    if not spellId then return end
    
    lib:Print("Ending cast")
    local mob = getMobDataForGUID(srcGUID)
    local speed = getMobPropertyForMob(mob, "attackspeedMain", srcGUID)
    
    local spellData = getSpellData(spellId)
    local casttimeData = getCastPropertyDataForSpell(spellData, "casttime")
    
    setEndToTimeProperty(casttimeData, timestamp - speed, srcGUID)
    self.castingSpell[srcGUID] = nil
    
    local affectsPlayerData = getCastPropertyDataForSpell(spellData, "affectsPlayer")
    addValueToBoolProperty(affectsPlayerData, false)
end

function SpellHandler:EvaluateSpellHit(timestamp, eventtype, srcGUID, srcName, dstGUID, dstName, spellId, spellName, hitType)
    if hitType == "ABSORB" then return end
    
    local hitData = getSpellPropertyDataForSpellID(spellId, "hitChance")
    
    if not hitData.statisticData then
        hitData.statisticData = {count = 1, sum = 1, forGUID = {}}
    end
    
    local hit = 0
    if hitType == "HIT" then
        hit = 1
    end   
    --lib:Print("Addsumtoprop: hitChanceMain")
    addSumDataToProperty(hitData, hit, dstGUID)
end

function SpellHandler:EvaluateSpellDamage(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, amount, school, resisted, blocked, absorbed, critical, glancing, crushing)
    local damageData = getSpellPropertyDataForSpellID(spellId, "amount")
    
    amount = amount + (absorbed or 0) + (resisted or 0)
    --lib:Print("Addsumtoprop: damageMain")
    addSumDataToProperty(damageData, amount)
    
    local factor = (critical and 0.5) or (crushing and 0.6666667) or 1.0
    damageData.guidValue[dstGUID] = amount * factor
end

function SpellHandler:OnSpellMiss(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, missType)
    self:EvaluateSpellHit(timestamp, eventtype, srcGUID, srcName, dstGUID, dstName, spellId, spellName, missType)
    
    local spellData = getSpellData(spellId)
    local affectsPlayerData = getCastPropertyDataForSpell(spellData, "affectsPlayer")
    addValueToBoolProperty(affectsPlayerData, true)
    
    local casttimeData = getCastPropertyDataForSpell(spellData, "casttime")
    
    setEndToTimeProperty(casttimeData, timestamp, srcGUID)
    
    lib:EventStopped(timestamp, spellName, nil, srcGUID, dstGUID, false, nil)
    setUnitsLastTarget(srcGUID, dstGUID)
    
    if hasTimePropertyStarted(casttimeData, srcGUID) then
        lib:Print("Spell already started again.")
        lib:EventStarted(spellName, timestamp, 
                         getSpellPropertyForSpell(spellData, "casttime", srcGUID), 
                         srcGUID, dstGUID, 
                         getSpellPropertyForSpell(spellData, "amount", dstGUID), 
                         getSpellPropertyForSpell(spellData, "hitChance", dstGUID))
    end
    
    self.castingSpell[srcGUID] = nil
end

function SpellHandler:OnSpellHit(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, amount, school, resisted, blocked, absorbed, critical, glancing, crushing)
    self:EvaluateSpellHit(timestamp, eventtype, srcGUID, srcName, dstGUID, dstName, spellId, spellName, "HIT")
    self:EvaluateSpellDamage(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, amount, school, resisted, blocked, absorbed, critical, glancing, crushing)
    
    local spellData = getSpellData(spellId)
    local affectsPlayerData = getCastPropertyDataForSpell(spellData, "affectsPlayer")
    addValueToBoolProperty(affectsPlayerData, true)
    
    local casttimeData = getCastPropertyDataForSpell(spellData, "casttime")
    
    setEndToTimeProperty(casttimeData, timestamp, srcGUID)
    
    lib:EventStopped(timestamp, spellName, nil, srcGUID, dstGUID, true, amount)
    
    setUnitsLastTarget(srcGUID, dstGUID)
    
    if hasTimePropertyStarted(casttimeData, srcGUID) then
        lib:Print("Spell already started again.")
        lib:EventStarted(spellName, timestamp, 
                         getSpellPropertyForSpell(spellData, "casttime", srcGUID), 
                         srcGUID, dstGUID, 
                         getSpellPropertyForSpell(spellData, "amount", dstGUID), 
                         getSpellPropertyForSpell(spellData, "hitChance", dstGUID))
    end
    
    self.castingSpell[srcGUID] = nil
end

---------------------------
-- Combat event handlers --
---------------------------
local timeStart = nil;
function lib:PrintEvent(time, mob, msg)
    if not timeStart then timeStart = time end
    lib:Print(("[|cffefefff%.3f|r][|cffffffef%d|r] - %s"):format(time-timeStart, MobID[mob], msg))
end

function lib:SwingMissed(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags,missType)
    lib:PrintEvent(timestamp, srcGUID, "Swing missed")
    if isPlayer(dstFlags) then 
        SwingHandler:OnMobMiss(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags,missType)
        SpellHandler:OnSwing(timestamp, srcGUID)
    elseif missType == "PARRY" then
        SwingHandler:OnMobParry(timestamp, dstGUID)
    end
end

function lib:SwingDamage(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags,amount, school, resisted, blocked, absorbed, critical, glancing, crushing)
    lib:PrintEvent(timestamp, srcGUID, "Swing hit")
    if isPlayer(dstFlags) then 
        SwingHandler:OnMobHit(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags,amount, school, resisted, blocked, absorbed, critical, glancing, crushing)
        SpellHandler:OnSwing(timestamp, srcGUID)
    end
end

function lib:UnitDead(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags)    
    if isMob(dstFlags) then
        lib:PrintEvent(timestamp, dstGUID, "Mob dead")
        self:DeathOfMob(timestamp, dstGUID)
    end
end


function lib:SpellDamage(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, ...)
    if srcName and isMob(srcFlags) then
        lib:PrintEvent(timestamp, srcGUID, "Spell damage")
        --self:EvaluateReplaceSwingEnd(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool)
        SpellHandler:OnSpellHit(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, ...)
    end
end

function lib:SpellMiss(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, ...)
    if srcName and isMob(srcFlags) then
        lib:PrintEvent(timestamp, srcGUID, "Spell miss")
        SpellHandler:OnSpellMiss(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, missType)
    end
end

function lib:SpellSuccess(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, ...)
    if srcName and isMob(srcFlags) then
        lib:PrintEvent(timestamp, srcGUID, "Spell success")
        
        --check if this spell happens at next swing, like cleave and heroic strike
        --self:EvaluateReplaceSwingStart(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool)
    end
end

function lib:SpellStart(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool)
    if srcName and isMob(srcFlags) then
        lib:PrintEvent(timestamp, srcGUID, "Spell start")
        SwingHandler:OnSpellStart(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool)
        SpellHandler:OnSpellStart(timestamp, srcGUID, spellId, spellName, spellSchool)
    end
end

--------------------
-- Event Handlers --
--------------------

local EventParse =
{
	["SWING_DAMAGE"] = lib.SwingDamage, 
	["RANGE_DAMAGE"] = lib.SwingDamage, 
	["SWING_MISSED"] = lib.SwingMissed, 
	["RANGE_MISSED"] = lib.SwingMissed,
    ["UNIT_DIED"] = lib.UnitDead,
    ["SPELL_DAMAGE"] = lib.SpellDamage,
    ["SPELL_MISSED"] = lib.SpellMiss,
    ["SPELL_CAST_START"] = lib.SpellStart,
    ["SPELL_CAST_SUCCESS"] = lib.SpellSuccess,
}

local lastCombatEvent = 0
local eventCount = 0

function lib:COMBAT_LOG_EVENT_UNFILTERED(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)    
    if eventCount < 0 then
        if timestamp < lastCombatEvent then
            return
        else
            eventCount = 0
        end
    end
    
    if not refTime then
        refTime = GetTime() - timestamp
    end
    
	local parsefunc = EventParse[eventtype]
    
	
	if parsefunc then
        if timestamp - lastCombatEvent > 0.2 then
            lastCombatEvent = timestamp
            eventCount = 0
        else
            if eventCount > 16 then 
                eventCount = -1
                lastCombatEvent = timestamp + 15
                lib:Print("To much events...disable tracking for 15 seconds!")
                return 
            end
            eventCount = eventCount + 1
        end
        self:Print("-------------------------------")
		parsefunc(self, timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
	else
        if srcName and srcFlags and isMob(srcFlags) then
            self:Print((srcName or "<nil>")..": Unknown combat log event type: "..eventtype)
        end
	end
end


local elapsedTimer = 0
function lib:OnUpdate(elapsed)
    elapsedTimer = elapsedTimer + elapsed
    if elapsedTimer > 0.25 then
        clearOutdatedData() --check for events we expected but which did not happen
        
        elapsedTimer = 0
        --if InCombatLockdown() then
            --lib:Print("...")
        --end
    end
end

function lib:PLAYER_ALIVE()
    -- This event is only fired at initial login, not at reloadui or load-on-demand loading.
    -- The initialisation is triggered again, since none of the initialisation had any effect
    -- prior to this event firing (no messages sent and InBattlegroundOrArena and InRaidOrParty
    -- are probably not correctly initialised).
    lib:Initialise();
end

function lib:PLAYER_ENTERING_WORLD()
    --cleanup
end

function lib:PARTY_MEMBERS_CHANGED()
    InRaidOrParty = (GetNumRaidMembers() > 0) or (GetNumPartyMembers() > 0)

end

function lib:RAID_ROSTER_UPDATE()
    self:PARTY_MEMBERS_CHANGED()
end

function lib:PLAYER_REGEN_ENABLED()
    SwingHandler:OnLeftCombat()
end

function lib:Enable()
    --lib.EventFrame:SetScript("OnUpdate", function (this, ...) lib.OnUpdate(lib, ...) end)
    lib.EventFrame:UnregisterAllEvents()

    -- Register Events
    lib.EventFrame:RegisterEvent("PLAYER_ALIVE")
    lib.EventFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")

    -- For keeping track of versions
    lib.EventFrame:RegisterEvent("PARTY_MEMBERS_CHANGED")
    lib.EventFrame:RegisterEvent("RAID_ROSTER_UPDATE")

    -- Prune data at zone change
    lib.EventFrame:RegisterEvent("PLAYER_ENTERING_WORLD")
    
    lib.EventFrame:RegisterEvent("PLAYER_REGEN_ENABLED")
end

function lib:Disable()
    lib.EventFrame:SetScript("OnUpdate", nil)
    lib.EventFrame:UnregisterAllEvents()
end

function lib:Initialise()
    self:Enable()
    local it = select(2, IsInInstance())
    InBattlegroundOrArena = (it == "pvp") or (it == "arena")

    InRaidOrParty = (GetNumRaidMembers() > 0) or (GetNumPartyMembers() > 0)

end

lib:Initialise()
